@@ -34,6 +34,7 @@ import warnings
 import errno
 import hashlib
 import collections
+import urllib
 
 import disdat.apply   # <--- imports api, which imports apply, etc.
 import disdat.run
@@ -41,6 +42,7 @@ import disdat.fs
 import disdat.common as common
 from disdat.pipe_base import PipeBase
 from disdat.data_context import DataContext
+from disdat.utility.aws_s3 import s3_path_exists
 
 from disdat.hyperframe import HyperFrameRecord, LineageRecord
 from disdat.run import Backend, run_entry
@@ -201,6 +203,12 @@ class Bundle(HyperFrameRecord):
 
     @property
     def data(self):
+        """ Return the presented data if closed (refreshing in case of localization / delocalization).
+        Or return the data currently attached to an open bundle (if not closed).
+        """
+        if self._closed:
+            assert self.is_presentable
+            self._data = self.data_context.present_hfr(self) # actualize link urls
         return self._data
 
     @property
@@ -259,9 +267,12 @@ class Bundle(HyperFrameRecord):
         found = {}
         arg_names = ['_arg_{}'.format(i) for i in range(0, len(self.pb.lineage.depends_on))]
         for an, dep in zip(arg_names, self.pb.lineage.depends_on):
-            if dep.HasField("arg_name"):
-                found[dep.arg_name] = (dep.hframe_proc_name, dep.hframe_uuid)
-            else:
+            try:
+                if dep.arg_name:
+                    found[dep.arg_name] = (dep.hframe_proc_name, dep.hframe_uuid)
+                else:
+                    found[an] = (dep.hframe_proc_name, dep.hframe_uuid)
+            except ValueError as ve:
                 found[an] = (dep.hframe_proc_name, dep.hframe_uuid)
         return found
 
@@ -463,6 +474,7 @@ class Bundle(HyperFrameRecord):
         self.pb = hfr.pb
         self.init_internal_state()
         self._data = self.data_context.present_hfr(hfr)
+        self._local_dir, self._remote_dir = self.data_context.util_get_managed_paths(self.uuid)
         return self
 
     """ Python Context Manager Interface """
@@ -521,7 +533,7 @@ class Bundle(HyperFrameRecord):
             if self.processing_name == '':
                 self.processing_name = self.default_processing_name()
             self._mod_finish(new_time=True)   # Set the hash based on all the context now in the pb, record create time
-            self.data_context.write_hframe(self)
+            self.data_context.write_hframe_local(self)
         except Exception as error:
             """ If we fail for any reason, remove bundle dir and raise """
             self.abandon()
@@ -559,9 +571,12 @@ class Bundle(HyperFrameRecord):
         self._fs.commit(None, None, uuid=self.uuid, data_context=self.data_context)
         return self
 
-    def push(self):
+    def push(self, delocalize=False):
         """ Shortcut version of api.push(uuid=bundle.uuid)
 
+        Args:
+            delocalize (bool): Remove all local files, default False
+
         Returns:
             (`disdat.api.Bundle`): this bundle
 
@@ -572,7 +587,12 @@ class Bundle(HyperFrameRecord):
             raise RuntimeError("Not pushing: Current branch '{}/{}' has no remote".format(self.data_context.get_remote_name(),
                                                                                           self.data_context.get_local_name()))
 
-        self._fs.push(uuid=self.uuid, data_context=self.data_context)
+        self._fs.push(uuid=self.uuid, data_context=self.data_context, delocalize=delocalize)
+
+        # if we removed files, update the the presentation
+        if delocalize:
+            assert self.is_presentable
+            self._data = self.data_context.present_hfr(self) # actualize link urls
 
         return self
 
@@ -584,20 +604,89 @@ class Bundle(HyperFrameRecord):
         """
         self._check_closed()
         self._fs.pull(uuid=self.uuid, localize=localize, data_context=self.data_context)
+
+        # if we pulled files, update the the presentation
         if localize:
             assert self.is_presentable
             self._data = self.data_context.present_hfr(self) # actualize link urls
         return self
 
+    def localize(self, path):
+        """ Given a remote managed path in a bundle, retrieve the file.
+
+        A local path should already be localized, but in case the user keeps a
+        local path around, double check that it is local.
+
+        Assumes the bundle has been pushed via b.push()
+
+        Args:
+            path (str): The managed s3 path
+
+        Returns:
+            str: The localized path
+
+        """
+        self._check_closed()
+        scheme = urllib.parse.urlparse(path).scheme
+        assert scheme == 's3', "Bundle.localize: Bad scheme {}, requires an s3 path".format(scheme)
+        assert self._remote_dir is not None, "Bundle.localize failed, bundle created in context without remote."
+        assert path.startswith(self._remote_dir), "Bundle.localize failed, {} is not remote managed path in {}".format(path, self._remote_dir)
+
+        # Pull the one file.
+        new_path = self.data_context.copy_in_files(path, urllib.parse.urljoin('file:', self._local_dir))
+
+        # Update the presentation
+        assert self.is_presentable
+        self._data = self.data_context.present_hfr(self)  # actualize link urls
+
+        return new_path
+
+    def delocalize(self, path):
+        """ Given a local managed path in a bundle, delete the local copy.
+
+        Tests that the metadata and this item have been pushed.
+
+        Args:
+            path: A managed local or s3 path.
+            force(bool):  Delete the file without checking if the bundle is at the current remote.
+
+        Returns:
+            str: The path to the non-localized file or None if force is set and no remote available.
+
+        """
+        self._check_closed()
+        scheme = urllib.parse.urlparse(path).scheme
+        assert scheme == '' or scheme == 'file', "Bundle.delocalize: Bad scheme {}, requires a local path".format(scheme)
+        assert self._remote_dir is not None, "Bundle.delocalize failed, bundle created in context without remote."
+        assert path.startswith(self._local_dir), "Bundle.delocalize failed, {} is not a managed path in {}".format(path, self._remote_dir)
+        hframe_file = os.path.join(self._remote_dir, "{}_hframe.pb".format(self.uuid))
+        assert s3_path_exists(hframe_file), "Bundle.delocalize failed. Bundle must be pushed to remote before use."
+
+        # Push the one file.   This is a safety call.
+        # If we can't guarantee that we've made a copy, we don't remove the local copy.
+        new_path = self.data_context.copy_in_files(urllib.parse.urljoin('file:', path), self._remote_dir)
+
+        # Delete local copy
+        try:
+            os.remove(path)
+        except IOError as e:
+            print("delocalize: unable to remove {} due to {}".format(path, e))
+
+        # Update the presentation
+        assert self.is_presentable
+        self._data = self.data_context.present_hfr(self)  # actualize link urls
+
+        return new_path
+
     def get_directory(self, dir_name):
         """ Returns path `<disdat-managed-directory>/<dir_name>`.  This gives the user a local
         output directory into which to write files.  This is useful when a user needs to give an external tool, such
-        as Spark or Tensorflow, a directory to place output files.
+        as Spark or Tensorflow, a directory to place output files.   After this call, the directory
+        will exist in the local context.
 
-        Add this path as you would add file paths to your output bundle.  Disdat will incorporate
-        all the data found in this directory into the bundle.
+        It is the user's responsibility to add individual file links to the bundle.
 
-        See Pipe.create_output_dir()
+        It is an error to add a directory as a file link.
 
         Arguments:
             dir_name (str): A basedir of a directory to appear in the local bundle.
@@ -622,7 +711,6 @@ class Bundle(HyperFrameRecord):
             if not why.errno == errno.EEXIST:
                 _logger.error("Creating directory in bundle directory failed errno {}".format(why.strerror))
                 raise
-            # else directory exists is OK and fall through
         except IOError as why:
             _logger.error("Creating directory in bundle directory failed {}".format(why))
             raise
@@ -634,8 +722,9 @@ class Bundle(HyperFrameRecord):
         This allows you to create versioned data sets without file copies and without worrying about where
         your data files are to be stored.
 
-        To use, you must a.) write data into this file-like object (a 'target'), and b.) add this
-        target to the bundle by either including its file path or the luigi target object itself.
+        To use, you must:
+            a.) Write data into this location
+            b.) Add this target to the bundle by including this file path
 
         TODO: for binary files add a binary=True, and return luigi.LocalTarget('test.npz', format=luigi.format.Nop)
 
@@ -643,20 +732,19 @@ class Bundle(HyperFrameRecord):
             filename (str,list,dict): filename to create in the bundle
 
         Returns:
-            `luigi.LocalTarget`
+            str
         """
         self._check_open()
-        return PipeBase.filename_to_luigi_targets(self._local_dir, filename)
+        return os.path.join(self._local_dir, filename)
 
     def get_remote_directory(self, dir_name):
         """ Returns path `<disdat-managed-remote-directory>/<dir_name>`.  This gives the user a remote (e.g., s3)
         output directory into which to write files.  This is useful when a user needs to give an external tool, such
         as Spark or Tensorflow, a directory to place output files.
 
-        Add this path as you would add file paths to your output bundle.  Disdat will incorporate
-        all the data found in this directory into the bundle.
+        It is the user's responsibility to add individual file links to the bundle.
 
-        See Pipe.create_output_dir_remote()
+        It is an error to add a directory as a file link.
 
         Arguments:
             dir_name (str): A basedir of a directory to appear in the remote bundle.
@@ -665,9 +753,7 @@ class Bundle(HyperFrameRecord):
             str: A directory path managed by disdat
         """
         self._check_open()
-
         fqp = os.path.join(self._remote_dir, dir_name.lstrip('/'))
-
         return fqp
 
     def get_remote_file(self, filename):
@@ -690,7 +776,7 @@ class Bundle(HyperFrameRecord):
         if not self.data_context.remote_ctxt_url:
             raise Exception('Managed S3 path creation requires a remote context')
 
-        return PipeBase.filename_to_luigi_targets(self._remote_dir, filename)
+        return os.path.join(self._remote_dir, filename)
 
     @staticmethod
     def calc_default_processing_name(code_ref, params, dep_proc_ids):
@@ -1012,7 +1098,7 @@ def get(local_context, bundle_name, uuid=None, tags=None):
     return b
 
 
-def rm(local_context, bundle_name=None, uuid=None, tags=None, rm_all=False, rm_old_only=False, force=False):
+def rm(local_context, bundle_name=None, uuid=None, tags=None, rm_all=False, rm_old_only=False):
     """ Delete a bundle with a certain name, uuid, or tags.  By default removes the most recent bundle.
     Otherwise one may specify `rm_all=True` to remove all bundles, or `rm_old_only=True` to remove
     all bundles but the most recent.
@@ -1024,7 +1110,6 @@ def rm(local_context, bundle_name=None, uuid=None, tags=None, rm_all=False, rm_o
         tags (dict(str:str)): Optional dictionary of tags that must be present on bundle to remove
         rm_all (bool): Remove latest and all historical if given bundle_name
         rm_old_only (bool): remove everything but latest if given bundle_name
-        force (bool): If a db-link exists and it is the latest on the remote DB, force remove. Default False.
 
     Returns:
         None
@@ -1035,7 +1120,7 @@ def rm(local_context, bundle_name=None, uuid=None, tags=None, rm_all=False, rm_o
     data_context = _get_context(local_context)
 
     fs.rm(human_name=bundle_name, rm_all=rm_all, rm_old_only=rm_old_only,
-          uuid=uuid, tags=tags, force=force, data_context=data_context)
+          uuid=uuid, tags=tags, data_context=data_context)
 
 
 def add(local_context, bundle_name, path, tags=None):
@@ -1152,14 +1237,15 @@ def commit(local_context, bundle_name, tags=None, uuid=None):
     fs.commit(bundle_name, tags, uuid=uuid, data_context=data_context)
 
 
-def push(local_context, bundle_name, tags=None, uuid=None):
+def push(local_context, bundle_name=None, tags=None, uuid=None, delocalize=False):
     """ Push a bundle to a remote repository.
 
     Args:
-        local_context (str):  The local context to push to (must have a remote).
+        local_context (str):  The local context to push to (must have a remote).  Default None means push all.
         bundle_name (str): human name of the bundle to push or None (if using uuid)
         tags (dict): Tags the bundle must have
         uuid (str): Optional UUID of the bundle to push.  UUID takes precedence over bundle name
+        delocalize (bool): After pushing, remove all local file links (default False)
 
     Returns:
         None
@@ -1174,7 +1260,7 @@ def push(local_context, bundle_name, tags=None, uuid=None):
         raise RuntimeError("Not pushing: Current branch '{}/{}' has no remote".format(data_context.get_remote_name(),
                                                                                       data_context.get_local_name()))
 
-    fs.push(human_name=bundle_name, uuid=uuid, tags=tags, data_context=data_context)
+    fs.push(human_name=bundle_name, uuid=uuid, tags=tags, data_context=data_context, delocalize=delocalize)
 
 
 def pull(local_context, bundle_name=None, uuid=None, localize=False):
@@ -1408,3 +1494,23 @@ def dockerize_get_id(setup_dir):
         (str): The full docker container image hash
     """
     return dockerize_entry(pipeline_root=setup_dir, get_id=True)
+
+
+def helper_get_files_in_dir(dir):
+    """ Return all files under this directory.
+    If no scheme, assume local files.
+    1.) Only look one-level down (in this directory)
+    2.) Do not include anything that looks like one of disdat's pbufs
+
+    TODO: One place that defines the format of the Disdat pb file names
+    See data_context.DataContext: rebuild_db() *_frame.pb, *_hframe.pb, *_auth.pb
+    Args:
+        (str): local directory
+    Returns:
+        (list:str): List of files in that directory
+    """
+
+    files = [os.path.join(dir, f) for f in os.listdir(dir) if os.path.isfile(os.path.join(dir, f))
+             and ('_hframe.pb' not in f) and ('_frame.pb' not in f) and ('_auth.pb' not in f)]
+
+    return files
